/*
 * filter.c: URL filter daemon
 *
 * Copyright (C) 2021 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Jomily Joseph <jomily.joseph@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>

#include <string.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <linux/netfilter.h>
#include <linux/udp.h>
#include <syslog.h>
#include <time.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
#include <libnetfilter_queue/linux_nfnetlink_queue.h>
#include <uci.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <pthread.h>

#include "config.h"
#include "filter.h"
#include "dns.h"

#define CHECK_GLOBAL_URL_LIST -1
#define URL_SUCCESS 1
#define MAX_IP_LEN 32

struct ubus_context *ctx;
struct nfq_handle *nfq_h;
struct nfq_q_handle *queue_h;
struct uloop_fd uloop_fd;
char **dns_server_ip;

int dns_server_cnt;


/* return 1 if IP string is valid, else return 0 */
int validate_ip(char *ip_str)
{
	struct sockaddr_in sa;

	return inet_pton(AF_INET, ip_str, &(sa.sin_addr));

}

/* return 1 if IPv6 string is valid, else return 0 */
int validate_ipv6(char *ip6_str)
{
	struct sockaddr_in6 sa6;

	return inet_pton(AF_INET6, ip6_str, &(sa6.sin6_addr));
}


static int is_uci_option_valid(struct uci_package *net_pkg, char *option, char *value)
{
	struct uci_element *element = NULL;
	struct uci_section *s = NULL;
	struct uci_option *o = NULL;
	struct uci_element *e = NULL;

	uci_foreach_element(&net_pkg->sections, e) {
		s = uci_to_section(e);

		if (strcmp(s->type, "interface") != 0)
			continue;

		uci_foreach_element(&s->options, element) {
			o = uci_to_option(element);
			if (o) {
				if (!(strcmp(option, "ip6addr") == 0)) {
					if ((strcmp(o->e.name, option) == 0) && (strcmp(o->v.string, value) == 0))
						return 1;
				} else {
					if ((strcmp(o->e.name, option) == 0) && (validate_ipv6(o->v.string)))
						return 1;
				}
			}
		}
	}

	return 0;
}


static int is_ipv6_enabled(void)
{
	static struct uci_context *net_ctx;
	static struct uci_package *net_pkg;
	int ipv6_proto_dhcpv6_valid, ipv6_proto_static_valid, ipv6_address_valid;

	net_ctx = uci_alloc_context();
	if (!net_ctx) {
		syslog(LOG_ERR, "failed to allocate uci context");
		return -1;
	}

	if (uci_load(net_ctx, "network", &net_pkg)) {
		syslog(LOG_ERR, "failed to load uci file");
		return -1;
	}
	ipv6_proto_dhcpv6_valid = is_uci_option_valid(net_pkg, "proto", "dhcpv6");
	ipv6_proto_static_valid = is_uci_option_valid(net_pkg, "proto", "static");
	ipv6_address_valid = is_uci_option_valid(net_pkg, "ip6addr", "");

	if (ipv6_proto_dhcpv6_valid || (ipv6_proto_static_valid && ipv6_address_valid))
		return 1;

	return 0;
}

// Function is used to get the payload from the packet.
// @param ipv4 is set if packet is ipv4
// @param data_p packet data
// @param proto protocol type
// @param len length of data
// @return on success returns payload offset else -1
static int get_ip_packet_payload(unsigned char *data_p, uint8_t *proto, int len, bool *is_ipv6)
{
	const struct iphdr *iph;
	const struct ip6_hdr *ip6h;
	const struct tcphdr *tcp_header;
	const struct udphdr *udp_header;
	int payload_offset;
	uint8_t ipv4;

	iph = (struct iphdr *)data_p;
	ipv4 = (iph->version == 4)?1:0;

	if (ipv4) {
		iph = (const struct iphdr *)data_p;
		*proto = iph->protocol;
		*is_ipv6 = false;

		if (*proto == IPPROTO_UDP) {
			// Check if the packet is dns packet or not
			udp_header = (const struct udphdr *)(data_p + (iph->ihl<<2));
			if (((ntohs(udp_header->dest)) != DNS_PORT) && ((ntohs(udp_header->source)) != DNS_PORT)) {
				syslog(LOG_ERR, "Not a dns packet");
				return -1;
			}

			// If the packet is dns response then return the payload_offset as
			// 0 and update the dns cache with the ip address.
			if (ntohs(udp_header->source) == DNS_PORT) {
				unsigned char dns_payload[len + 1];

				memset(dns_payload, '\0', len + 1);
				memcpy(dns_payload, data_p+((iph->ihl)<<2) + sizeof(struct udphdr), len-(((iph->ihl)<<2) + sizeof(struct udphdr)) + 1);
				if ('\0' == dns_payload[0]) {
					syslog(LOG_ERR, "payload is null");
					return -1;
				}
				parse_dns_response(dns_payload);
				return 0;
			}

			// If the packet is dns query then return the payload offset.
			payload_offset = ((iph->ihl)<<2) + sizeof(struct udphdr) + sizeof(struct dnshdr);

		} else if (*proto == IPPROTO_TCP) {
			// Return the payload offset in case of tcp packet.
			tcp_header = (const struct tcphdr *)(data_p + (iph->ihl<<2));
			payload_offset = ((iph->ihl)<<2) + (tcp_header->doff<<2);
		} else {
			syslog(LOG_ERR, "Unknown protocol");
			return -1;
		}

		return payload_offset;
	} else {
		ip6h = (struct ip6_hdr *)data_p;
		//ipv6 handling
		if ((ip6h->ip6_vfc >> 4) != 6)
			return -1;
		*proto = ip6h->ip6_nxt;
		*is_ipv6 = true;

		payload_offset = ipv6_header_get_L4_header_offset(ip6h, proto);

		if (*proto == IPPROTO_UDP) {
			// Check if the packet is dns packet or not
			udp_header = (const struct udphdr *)((uint8_t *)ip6h + payload_offset);
			if (((ntohs(udp_header->dest)) != DNS_PORT) && ((ntohs(udp_header->source)) != DNS_PORT)) {
				syslog(LOG_ERR, "Not a dns packet");
				return -1;
			}

			// Don't process DNS packets if ipv6 is not enabled
			if (!is_ipv6_enabled())
				return 1;
			// If the packet is dns response then return the payload_offset as
			// 0 and update the dns cache with the ip address.
			if (ntohs(udp_header->source) == DNS_PORT) {
				unsigned char dns_payload[len + 1];

				memset(dns_payload, '\0', len + 1);
				memcpy(dns_payload, data_p+payload_offset+sizeof(struct udphdr), len-(payload_offset+sizeof(struct udphdr)) + 1);
				if ('\0' == dns_payload[0]) {
					syslog(LOG_ERR, "payload is null");
					return -1;
				}
				parse_dns_response(dns_payload);
				return 0;
			}

			// If the packet is dns query then return the payload offset.
			payload_offset += sizeof(struct udphdr) + sizeof(struct dnshdr);
		} else if (*proto == IPPROTO_TCP) {
			tcp_header = (const struct tcphdr *)((uint8_t *)ip6h + payload_offset);
			payload_offset += (tcp_header->doff<<2);
		} else {
			syslog(LOG_ERR, "Unknown protocol");
			return -1;
		}

		return payload_offset;
	}

	return -1;
}

// Function checks if the URL requested is globally blacklisted or not.
// @param payload_pkt packet payload
// @param url_found indicates if url was present in global list or not
// @return PKT_ACCEPT or PKT_DROP
static int global_url_filtering(char *payload_pkt, int *url_found)
{
	int i = 0;

	// Check if global blacklisting is enabled or not.
	if (true != url_global.enable_global_blacklisting) {
		syslog(LOG_INFO, "global blacklisting not enabled.");
		return PKT_ACCEPT;
	}

	// Check if the url is present in the global blacklist urls,
	// then drop the packet else accept the packet.
	for (i = 0; i < url_global.blacklist_url_cnt; i++) {
		char url[512] = "";

		strncpy(url, url_global.blacklist_url[i], sizeof(url));

		// sometimes name do not contain the .com but some other
		// extension, so make the match more effective, just
		// compare the domain name here
		char *token = strtok(url_global.blacklist_url[i], ".");

		if (token != NULL)
			strncpy(url, token, sizeof(url));

		if (strstr(payload_pkt, url) != NULL)
			return PKT_DROP;
	}

	*url_found = 0;
	return PKT_ACCEPT;
}

// Function checks if the URL requested is in blacklist or whitelist for a
// particular profile.
// @param payload_pkt packet payload
// @param match_profile profile store which has the requested mac address.
// @return PKT_ACCEPT or PKT DROP on success else -1.
static int match_url_from_profile(char *payload_pkt, profile *match_profile)
{
	int i;

	// Now check if the URL of the received packet is present in the whitelist url
	// of the above stored profile, if yes then accept the packet.
	for (i = 0; i < match_profile->whitelist_url_cnt; i++) {
		if (strstr(payload_pkt, match_profile->whitelist_url[i]) != NULL)
			return PKT_ACCEPT;
	}

	// Check if the URL of the received packet is present in the blacklist url of
	// the above stored profile, if yes then drop the packet.
	for (i = 0; i < match_profile->blacklist_url_cnt; i++) {
		if (strstr(payload_pkt, match_profile->blacklist_url[i]) != NULL)
			return PKT_DROP;
	}

	return CHECK_GLOBAL_URL_LIST;
}

// Fetches the current day and time
// @param day_index index of the day of week
// @return current time in seconds.
static unsigned int get_current_day_and_time(unsigned int *day_index)
{
	const char *weekdays[] = {"Sunday",
				"Monday", "Tuesday",
				"Wednesday", "Thursday",
				"Friday", "Saturday"};
	char day[MAX_LEN_WEEKDAY];
	time_t cur_time;
	struct tm *loc_time;
	int i = 0;

	// First fetch the day
	memset(day, '\0', MAX_LEN_WEEKDAY);
	cur_time = time(NULL);
	loc_time = localtime(&cur_time);
	strftime(day, MAX_LEN_WEEKDAY, "%A", loc_time);
	day[strcspn(day, "\r\n")] = 0;

	// Fetch index for the day.
	for (i = 0; i < MAX_DAYS; i++) {
		if (strncasecmp(day, weekdays[i], sizeof(day)) == 0) {
			*day_index = i;
			break;
		}
	}

	// Get the current time in seconds format.
	return ((loc_time->tm_hour * 60) + loc_time->tm_min) * 60 + loc_time->tm_sec;
}

// Function checks if the current day and time is within
// the given constrainst or not.
// @param i index of url_filter structure.
// @param day_index current day index.
// @param prev_day_index previous day index
// @param time_data structure storing data.
// @return 0 on success else -1.
static int check_if_time_within_range(int i, unsigned int day_index,
		unsigned int prev_day_index, time_info time_data)
{
	int j = 0;
	bool is_day_present = 0;

	// Check if the day is specified in the configuration.
	// If day is not specified in the filter config then the filter
	// will be applied for all days for the specified time and
	// duration.
	for (j = 0; j < MAX_DAYS; j++) {
		if (FILTER_ACTIVE == (url_filter + i)->day[j]) {
			is_day_present = 1;
			break;
		}
	}

	if (is_day_present == 0) {
		if ((time_data.current_time >= time_data.start_time) &&
				(time_data.current_time <= time_data.end_time))
			return 0;

		if (time_data.is_next_day == true) {
			if (time_data.current_time <= time_data.next_day_end_time)
				return 0;
		}

		return CHECK_GLOBAL_URL_LIST;
	}

	// Now check if the filter is active for the day if yes
	// then check if is_next_day_slot is enabled and the filter
	// was active for previous day. if yes then check the
	// time range in both current day slot and next day slot.
	// Else check only in current_day_time_slot.
	if ((url_filter + i)->day[day_index] == FILTER_ACTIVE) {
		if ((time_data.current_time >= time_data.start_time) &&
				(time_data.current_time <= time_data.end_time))
			return 0;

		if (time_data.is_next_day == true &&
				(url_filter + i)->day[prev_day_index] == FILTER_ACTIVE) {
			if (time_data.current_time <= time_data.next_day_end_time)
				return 0;
		}
	} else {
		// Now if the filter is not active for the day check if
		// is_next_day is enabled and the filter was active
		// for yesterday. If yes, then check the time slot from
		// next_day_time_slot.
		if (time_data.is_next_day == true &&
				(url_filter + i)->day[prev_day_index] == FILTER_ACTIVE) {
			if (time_data.current_time <= time_data.next_day_end_time)
				return 0;
		}
	}

	return CHECK_GLOBAL_URL_LIST;
}

// Function filters the packet based on time slots.
// @param i index of url_filter structure.
// @return 0 on success else -1.
static int time_based_url_filtering(int i)
{
	time_info time_data;
	unsigned int day_index = 0, prev_day_index = 0,
		     duration = 0, hh = 0, ss = 0, mm = 0;

	memset(&time_data, 0, sizeof(time_info));
	// Fetch the current day and time.
	time_data.current_time = get_current_day_and_time(&day_index);

	// Get previous day index
	prev_day_index = day_index ? day_index - 1 : MAX_DAYS;

	// Check if duration is greater than 24 hrs ie 86400 seconds if yes then
	// set duration to 24 hrs.
	duration = (url_filter + i)->duration;
	if (duration >= MAX_SEC)
		duration = MAX_SEC;

	// Convert start_time into seconds and add duration to it to get the end time.
	// If end time is greater than 86400 then enable is_next_day flag and calculate the
	// next day end time.
	sscanf((url_filter + i)->start_time, "%u:%u:%u", &hh, &mm, &ss);
	time_data.start_time = ((hh * 60) + mm) * 60 + ss;
	time_data.end_time = time_data.start_time + duration;
	if (time_data.end_time > MAX_SEC) {
		time_data.is_next_day = 1;
		time_data.next_day_end_time = time_data.end_time - MAX_SEC;
		time_data.end_time = MAX_SEC;
	}

	return check_if_time_within_range(i, day_index, prev_day_index, time_data);
}

// Find the url corresponding to the cname from the dns cache
// @param c_name cname
// @return url on success else NULL.
static char *match_cname_from_dns_cache(char *c_name)
{
	int i = 0, j = 0;

	for (i = 0; i < dns_cache_cnt; i++) {
		for (j = 0; j < (global_dns_cache + i)->cname_cnt; j++) {
			if (strstr(c_name, (global_dns_cache + i)->cname[j]) != NULL)
				return (global_dns_cache + i)->url;
		}
	}
	return NULL;
}

// Find the url corresponding to the IP address from the dns cache
// @param payload_pkt ip address
// @return url on success else NULL.
static char *match_ip_address_from_dns_cache(char *ip_address, bool is_ipv6)
{
	int i = 0, j = 0;

	for (i = 0; i < dns_cache_cnt; i++) {
		if (!is_ipv6) {
			for (j = 0; j < (global_dns_cache + i)->ip_addr_cnt; j++) {
				if (strcmp((global_dns_cache + i)->ip_addr[j], ip_address) == 0)
					return (global_dns_cache + i)->url;
			}
		} else {
			for (j = 0; j < (global_dns_cache + i)->ipv6_addr_cnt; j++) {
				if (strcmp((global_dns_cache + i)->ipv6_addr[j], ip_address) == 0)
					return (global_dns_cache + i)->url;
			}
		}
	}
	return NULL;
}

// Function checks if the requested URL is present in the profile.
// If yes, then either accepts it or drops it as per the configuration.
// @param payload_pkt packet data
// @param match_profile profile which matches with the source mac address
// @param is_https_packet https packet or not
// @return on success PKT_ACCEPT or PKT_DROP else -1.
static int mac_based_url_filtering(char *payload_pkt, profile *match_profile,
						bool is_https_packet, bool is_ipv6)
{
	int ret = CHECK_GLOBAL_URL_LIST;
	char *url_pkt = NULL;

	// If the packet is https then find the URL corresponding to
	// the IP address present in the payload_pkt from the dns cache.
	if (is_https_packet == true) {
		url_pkt = match_ip_address_from_dns_cache(payload_pkt, is_ipv6);
		if (url_pkt == NULL)
			return PKT_ACCEPT;
		ret = match_url_from_profile(url_pkt, match_profile);
	} else {
		ret = match_url_from_profile(payload_pkt, match_profile);
		if (ret == CHECK_GLOBAL_URL_LIST) {
			// Check if the url is present in cname list of dns cache.
			// If yes, then find the url name corresponding to the cname and check
			// if that is present in the matched profile. If yes, then drop the packet.
			char *cname = match_cname_from_dns_cache(payload_pkt);
			if (cname != NULL)
				ret = match_url_from_profile(cname, match_profile);
		}
	}

	return ret;
}

// Function finds the profile associated with the MAC address of the
// packet received and maps the URL with the requested URL.
// @param mac_addr mac address to be searched in the profile
// @param len length of mac_addr
// @return on success PKT_ACEEPT or PKT_DROP else -1.
static int profile_based_url_filtering(char *payload_pkt, char *mac_addr,
						int len, bool is_https_packet, bool is_ipv6)
{
	int i = 0, j = 0, ret = CHECK_GLOBAL_URL_LIST;
	profile *match_profile;

	// First check if this mac address is associated with any profile
	// in the filter configuration.
	for (i = 0; i < filter_cnt; i++) {
		// Check if the profile is enabled or not.
		if (((url_filter + i)->enable) == false)
			continue;

		// If mac address is found, we will check if the URL of the packet
		// received matches the one in the profile. Here we are assuming that
		// mac addresses are not duplicated in filters, if so, we use the
		// first profile in which we find the mac address.
		for (j = 0; j < (url_filter + i)->mac_cnt; j++) {
			if (strncasecmp(mac_addr, (url_filter + i)->mac_list[j], len) == 0) {
				match_profile = (url_filter + i)->p;
				ret = mac_based_url_filtering(payload_pkt,
						match_profile, is_https_packet, is_ipv6);

				// If the return value is -1 or time based url filtering
				// is not enabled then return the value from the
				// previous function.
				if (ret == CHECK_GLOBAL_URL_LIST ||
						(url_filter + i)->start_time[0] == '\0' ||
						(url_filter + i)->duration == 0) {
					return ret;
				}

				// If the packet falls between the time slot then
				// return either PKT_ACCEPT or PKT_DROP based on the
				// previous function else return -1.
				if (!time_based_url_filtering(i))
					return ret;
				else
					return CHECK_GLOBAL_URL_LIST;
			}
		}
	}

	return ret;
}

// Function gets the qname ie the hostname for which the dns query
// has been generated, from the packet
// @param data payload
// @domaaain_name variable to store the url
// @return nothing
char *get_domain_name(char data[])
{
	// The content of the QNAME is of the format 3www8mydomain3com0
	// instead of www.mydomain.com. Also this is a variable length
	// field to fit the hostname. So we need to parse the data to
	// get the url for which the dns query has been initiated.
	int i = 0, j = 0, start = 0;
	int toread = data[0];

	i++;

	while (toread != 0) {
		for (; i <= start+toread; i++) {
			data[j] = data[i];
			j++;
		}

		// Get next chunk
		data[j] = '.';
		j++;
		toread = data[i];
		start = i;
		i++;
	}

	data[i - 2] = '\0';
	return data;
}

// Check if the packet is https or not. If yes then,
// returns the destination IP address.
// @param data payload data
// @return On success destination ip address else NULL
static char *check_https_packet(unsigned char *data, char **payload_pkt, bool is_ipv6)
{
	struct sockaddr_in dest;
	const struct iphdr *iph;
	const struct tcphdr *tcp_header;
	uint8_t proto;
	const struct ip6_hdr *ip6h;
	struct sockaddr_in6 dest6;

	if (!is_ipv6) {
		iph = (const struct iphdr *)data;
		tcp_header = (const struct tcphdr *)(data + (iph->ihl<<2));

		memset(&dest, 0, sizeof(dest));
		dest.sin_addr.s_addr = iph->daddr;

		// Check if destination port is 443.
		if (ntohs(tcp_header->dest) == HTTPS_PORT)
			return inet_ntoa(dest.sin_addr);
	} else {
		ip6h = (const struct ip6_hdr *)data;
		tcp_header = (const struct tcphdr *)(data + ipv6_header_get_L4_header_offset((struct ip6_hdr *)data, &proto));

		memset(&dest6, 0, sizeof(dest6));
		memcpy(&dest6.sin6_addr.__in6_union.__s6_addr, &ip6h->ip6_dst, sizeof(struct in6_addr));

		if (ntohs(tcp_header->dest) == HTTPS_PORT)
			return inet_ntop(AF_INET6, &dest6.sin6_addr, *payload_pkt, MAX_IPV6_LEN);
	}

	return NULL;
}

// Function fetches the packet payload ie URL for http packet,
// destination IP address for https packet and domain name for
// dns packet
// @param proto protocol type
// @param data payload
// @param payload_offset data offset
// @param payload_pkt variable to store payload
// @param is_https_packet variable to store if its a https packet
// @return 1 or PKT_ACCEPT
static int get_ipv6_tcp_payload_pkt(unsigned char *data, char *payload_pkt, bool *is_https_packet)
{
	// Check if its https packet or not, if yes then update the
	// payload_pkt with destination ip address.
	payload_pkt = check_https_packet(data, &payload_pkt, true);
	if (payload_pkt != NULL && payload_pkt[0] != '\0') {
		*is_https_packet = true;
		return URL_SUCCESS;
	}
	return PKT_ACCEPT;
}

// Function fetches the packet payload ie URL for http packet,
// destination IP address for https packet and domain name for
// dns packet
// @param proto protocol type
// @param data payload
// @param payload_offset data offset
// @param payload_pkt variable to store payload
// @param is_https_packet variable to store if its a https packet
// @return 1 or PKT_ACCEPT
static int get_payload_pkt(int proto, unsigned char *data, int payload_offset,
		char **payload_pkt, bool *is_https_packet, bool is_ipv6)
{

	if (proto == IPPROTO_TCP) {
		// Check if its https packet or not, if yes then update the
		// payload_pkt with destination ip address.
		if (!is_ipv6) {
			*payload_pkt = check_https_packet(data, payload_pkt, is_ipv6);
			if (*payload_pkt != NULL && *payload_pkt[0] != '\0') {
				*is_https_packet = true;
				return URL_SUCCESS;
			}
		}

		// Since its not an ttps packet, check if it is an http packet or not
		*payload_pkt = (char *)(data + payload_offset);
		if (strstr(*payload_pkt, "GET ") == NULL  &&
				strstr(*payload_pkt, "POST ") == NULL &&
				strstr(*payload_pkt, "HEAD ") == NULL) {
			syslog(LOG_INFO, "No http/https packet");
			return PKT_ACCEPT;
		}
	} else if (proto == IPPROTO_UDP) {
		// If the packet is a dns reponse then accept the packet.
		if (payload_offset == 0)
			return PKT_ACCEPT;

		// Get the qname(hostname) from the packet.
		*payload_pkt = get_domain_name((char *)(data + payload_offset));
		if (*payload_pkt == NULL) {
			syslog(LOG_INFO, "Payload for dns packet is NULL");
			return PKT_ACCEPT;
		}
	}

	return URL_SUCCESS;
}

// Checks if the packet needs to be accepted or not based on mac address
// and url.
// @param payload packet payload
// @return PKT_ACCEPT or PKT_DROP
static int handle_packet_filtering(struct nfq_data *payload)
{
	struct nfqnl_msg_packet_hw *hw_hdr;
	unsigned char *data = NULL;
	char *payload_pkt = NULL;
	uint8_t proto;
	int data_len, ret = 0;
	bool is_https_packet = false;
	bool is_ipv6 = false;
	char ipv6_addr_str[MAX_IPV6_LEN];

	// Get the source mac address.
	hw_hdr = nfq_get_packet_hw(payload);
	if (!hw_hdr) {
		syslog(LOG_ERR, "Error in fetching mac");
		return PKT_ACCEPT;
	}

	char mac_addr[MAC_LEN];

	memset(mac_addr, '\0', MAC_LEN);
	snprintf(mac_addr, sizeof(mac_addr), "%02x:%02x:%02x:%02x:%02x:%02x",
			hw_hdr->hw_addr[0], hw_hdr->hw_addr[1], hw_hdr->hw_addr[2],
			hw_hdr->hw_addr[3], hw_hdr->hw_addr[4], hw_hdr->hw_addr[5]);

	if (mac_addr[0] == '\0') {
		syslog(LOG_ERR, "MAC address empty");
		return PKT_ACCEPT;
	}

	data_len = nfq_get_payload(payload, &data);
	if (data_len <= 0) {
		syslog(LOG_ERR, "No payload found");
		return PKT_ACCEPT;
	}

	// Get the payload offset.
	int payload_offset;

	payload_offset = get_ip_packet_payload(data, &proto, data_len, &is_ipv6);
	// Drop DNS oackets if ipv6 is not enabled
	if (payload_offset == 1) {
		syslog(LOG_INFO, "ipv6 is not enabled so drop ipv6 DNS packets");
		return PKT_DROP;
	}

	if (payload_offset < 0) {
		// Always accept the packet if error happens.
		syslog(LOG_INFO, "Error in payload offset");
		return PKT_ACCEPT;
	}

	// Get the URL(http)/Destination IP address(https)/Domain name(DNS)
	if ((proto == IPPROTO_TCP) && is_ipv6) {
		memset(ipv6_addr_str, 0, MAX_IPV6_LEN);
		ret = get_ipv6_tcp_payload_pkt(data, ipv6_addr_str, &is_https_packet);

		if (ret == URL_SUCCESS)
			payload_pkt = ipv6_addr_str;
	}
	if (ret != URL_SUCCESS) {
		ret = get_payload_pkt(proto, data, payload_offset,
				&payload_pkt, &is_https_packet, is_ipv6);
	}
	if (ret != URL_SUCCESS)
		return ret;

	if (payload_pkt == NULL) {
		syslog(LOG_ERR, "payload pkt NULL");
		return PKT_ACCEPT;
	}

	int len = strlen(payload_pkt);

	if (len <= 0) {
		syslog(LOG_ERR, "payload len less than 0");
		return PKT_ACCEPT;
	}

	// Check if the URL has been blacklisted in any profile.
	ret = profile_based_url_filtering(payload_pkt, mac_addr, sizeof(mac_addr), is_https_packet, is_ipv6);
	if (ret == CHECK_GLOBAL_URL_LIST) {
		// The destination (url or ip) is not blacklisted in a profile, check if the URL has been
		// globally blacklisted or not. If yes then drop the packet else accept the packet.
		char *url_name = NULL;
		int url_found = 1;

		if (is_https_packet == true) {
			url_name = match_ip_address_from_dns_cache(payload_pkt, is_ipv6);

			if (url_name == NULL) {
				// destination ip does not belong to a blacklisted url
				return PKT_ACCEPT;
			}

			ret = global_url_filtering(url_name, &url_found);
		} else {
			ret = global_url_filtering(payload_pkt, &url_found);
			if (url_found == 0) {
				// If yes, then find the url name corresponding to the cname and check
				// if that is present in the global list. If yes, then drop the packet.
				char *cname = match_cname_from_dns_cache(payload_pkt);
				if (cname  != NULL)
					ret = global_url_filtering(cname, &url_found);
			}
		}
	}

	return ret;
}


// Callback function for handling packets.
// @param queue_h queue 0 handle
// @param nfmsg nfq message received
// @param nfa message data
// @param data store data
// @return 0 on success and non zero otherwise
static int handle_packets(struct nfq_q_handle *queue_h, struct nfgenmsg *nfmsg,
		struct nfq_data *nfa, void *data)
{
	struct nfqnl_msg_packet_hdr *pkt_hdr;
	int pkt_decision, pkt_id = 0;

	// Fetch the packet header.
	pkt_hdr = nfq_get_msg_packet_hdr(nfa);
	if (pkt_hdr)
		pkt_id = ntohl(pkt_hdr->packet_id);

	// Check if we should block this packet on the basis of URL and source mac address.
	pkt_decision = handle_packet_filtering(nfa);
	if (pkt_decision == PKT_ACCEPT)
		return nfq_set_verdict(queue_h, pkt_id, NF_ACCEPT, 0, NULL);
	else
		return nfq_set_verdict(queue_h, pkt_id, NF_DROP, 0, NULL);
}

// Function to unbind and bind to nf_queue handler.
// @return nothing
static void update_nf_queue_handler(void)
{
	int nfq_bind_v4 = 1, nfq_bind_v6 = 1;

	// Unbinding existing nf_queue handler.
	if (nfq_unbind_pf(nfq_h, AF_INET) < 0) {
		syslog(LOG_ERR, "Error during nfq_unbind_pf() for IPv4");
		nfq_bind_v4 = 0;
	}

	if (nfq_unbind_pf(nfq_h, AF_INET6) < 0) {
		syslog(LOG_ERR, "Error during nfq_unbind_pf() for IPv6");
		nfq_bind_v6 = 0;
	}

	if (!(nfq_bind_v4 || nfq_bind_v6)) {
		syslog(LOG_ERR, "Error during nfq_unbind_pf()");
		exit(1);
	}

	// Binding nfnetlink_queue as nf_queue handler.
	nfq_bind_v4 = nfq_bind_v6 = 1;
	if (nfq_bind_pf(nfq_h, AF_INET < 0)) {
		syslog(LOG_ERR, "Error during nfq_bind_pf() for IPv4");
		nfq_bind_v4 = 0;
	}

	if (nfq_bind_pf(nfq_h, AF_INET6) < 0) {
		syslog(LOG_ERR, "Error during nfq_bind_pf() for IPv6");
		nfq_bind_v6 = 0;
	}

	if (!(nfq_bind_v4 || nfq_bind_v6)) {
		syslog(LOG_ERR, "Error during nfq_bind_pf()");
		exit(1);
	}
}

// Open a netlink connection and returns file descriptor
// @return fd of the netlink socket
static int open_netlink_connection(void)
{
	// Opening netlink socket.
	nfq_h = nfq_open();
	if (!nfq_h) {
		syslog(LOG_ERR, "Error during nfq_open()");
		exit(1);
	}

	update_nf_queue_handler();

	// Binding this socket to queue.
	queue_h = nfq_create_queue(nfq_h,  0, &handle_packets, NULL);
	if (!queue_h) {
		syslog(LOG_ERR, "Error during nfq_create_queue()");
		exit(1);
	}

	// Setting copy_packet mode.
	if (nfq_set_mode(queue_h, NFQNL_COPY_PACKET, 0xffff) < 0) {
		syslog(LOG_ERR, "nfq_set_mode error");
		exit(1);
	}

	return nfq_fd(nfq_h);
}

// This function prints the usage of the daemon
static void usage(char *prog)
{
	fprintf(stderr, "Usage: %s [options]\n", prog);
	fprintf(stderr, "\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "    -s <socket path>   ubus socket\n");
	fprintf(stderr, "    -h <help>       To print this help menu\n");
	fprintf(stderr, "\n");
}

// This function reads the packets received on the fd
// @param fd is the pointer to the file descriptor to read on
static void url_filter_recv(struct uloop_fd *fd)
{
	char buf[BUFSIZE];

	memset(buf, 0, sizeof(buf));
	while (1) {
		int rv = recv(fd->fd, buf, sizeof(buf), 0);
		if (rv >= 0) {
			nfq_handle_packet(nfq_h, buf, rv);
			memset(buf, 0, sizeof(buf));
		}
	}
}


// Get number of IPv4/v6 dns severs and IPv4/v6 addresses.
// @param filename name of resolver file
// @param store variable whether to store the IPv4/v6 address in global array.
static int fetch_nameservers(char *filename, bool store, int dns_list_offset, bool is_ipv6)
{
	int count = 0;
	char buffer[BUFSIZE];
	int ip_addr_len;
	FILE *fp;

	memset(buffer, '\0', BUFSIZE);

	if (filename == NULL || filename[0] == '\0') {
		syslog(LOG_ERR, "filename is null.");
		return -1;
	}

	fp = fopen(filename, "r");
	if (fp == NULL) {
		syslog(LOG_ERR, "File pointer empty");
		return -1;
	}

	if (is_ipv6)
		ip_addr_len = MAX_IPV6_LEN;
	else
		ip_addr_len = MAX_IP_LEN;

	while (fgets(buffer, sizeof(buffer), fp)) {
		// Remove trailing newline
		buffer[strcspn(buffer, "\n\r")] = 0;

		char *token, *end;

		token = strtok_r(buffer, " ", &end);
		while (token != NULL) {
			if (strncasecmp(token, "nameserver", 10) == 0) {
				if ((is_ipv6 && 1 == validate_ipv6(end)) || (!is_ipv6 && 1 == validate_ip(end))) {
					// Store the ip address to global array
					if (true == store)
						strncpy(dns_server_ip[count+dns_list_offset], end, ip_addr_len);
					count++;
				}
			}
			token = strtok_r(NULL, " ", &end);
		}
	}

	fclose(fp);

	return count;
}

// Get the name of resolver file from the dhcp uci file
// @param filename variable to store filename
// @return 0 on success otherwise -1
static int get_resolver_file(char *filename)
{

	static struct uci_context *dhcp_ctx;
	static struct uci_package *dhcp_pkg;
	struct uci_element *element = NULL;
	struct uci_element *e = NULL;
	struct uci_option *option = NULL;
	struct uci_section *s = NULL;


	// Read the resolve file from dhcp config.
	dhcp_ctx = uci_alloc_context();
	if (!dhcp_ctx) {
		syslog(LOG_ERR, "failed to allocate uci context");
		return -1;
	}

	if (uci_load(dhcp_ctx, "dhcp", &dhcp_pkg)) {
		syslog(LOG_ERR, "failed to load uci file");
		return -1;
	}

	uci_foreach_element(&dhcp_pkg->sections, e) {
		s = uci_to_section(e);

		// read only filter section, skip the rest
		if (strcmp(s->type, "dnsmasq") != 0)
			continue;

		uci_foreach_element(&s->options, element) {
			option = uci_to_option(element);
			if (strcmp(option->e.name, "resolvfile") == 0) {
				strncpy(filename, option->v.string, 256);
				return 0;
			}
		}
	}

	return -1;
}

// Create dns server ip list
// @return 0 on success else -1
static int init_dns_server_list(void)
{

	char filename[256] = {0};

	if (!get_resolver_file(&filename[0])) {
		// Get the count of nameservers corresponding to IPv4 address.
		int server_num = fetch_nameservers(filename, 0, 0, 0);

		if (server_num <= 0)
			syslog(LOG_ERR, "No nameservers found");



		// Get the count of nameservers corresponding to IPv6 address.
		int ipv6_server_num = fetch_nameservers(filename, 0, server_num, 1);

		if (ipv6_server_num <= 0)
			syslog(LOG_ERR, "No ipv6 nameservers found");

		if (ipv6_server_num || server_num) {
			// Create memory for storing the addresses.
			dns_server_ip = (char **)calloc(server_num+ipv6_server_num, sizeof(char *));
			if (dns_server_ip == NULL) {
				syslog(LOG_ERR, "Memory allocation failed for dns_server_ip");
				return -1;
			}

			for (int i = 0; i < server_num; i++) {
				dns_server_ip[i] = (char *)calloc(MAX_IP_LEN, sizeof(char));
				if (dns_server_ip[i] == NULL) {
					syslog(LOG_ERR, "Memory allocation failed for dns_server_ip");
					if (dns_server_ip != NULL)
						free(dns_server_ip);
					return -1;
				}
			}

			//Avoid fetching ipv6 name servers if ipv6 is not enabled
			if (!is_ipv6_enabled())
				ipv6_server_num = 0;

			for (int i = 0; i < ipv6_server_num; i++) {
				dns_server_ip[i+server_num] = (char *)calloc(MAX_IPV6_LEN, sizeof(char));
				if (dns_server_ip[i+server_num] == NULL) {
					syslog(LOG_ERR, "Memory allocation failed for ipv6 dns_server_ip");
					if (dns_server_ip != NULL)
						free(dns_server_ip);
					return -1;
				}
			}
			// Store the address.
			if (ipv6_server_num) {
				if (fetch_nameservers(filename, 1, server_num, 1) <= 0) {
					syslog(LOG_ERR, "reading dns server failed");
					return -1;
				}
			}

			// Store the address.
			if (server_num) {
				if (fetch_nameservers(filename, 1, 0, 0) <= 0) {
					syslog(LOG_ERR, "reading dns server failed");
					return -1;
				}
			}

			// Update the dns server count
			dns_server_cnt = server_num + ipv6_server_num;

		}

	}


	return 0;
}

int main(int argc, char **argv)
{
	const char *ubus_socket = NULL;
	int ch;

	while ((ch = getopt(argc, argv, "hs:")) != -1) {
		switch (ch) {
		case 's':
			ubus_socket = optarg;
			break;
		case 'h':
			usage(argv[0]);
			exit(0);
		default:
			break;
				}
	}

	openlog("urlfilter", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	uloop_init();
	ctx = ubus_connect(ubus_socket);
	if (!ctx) {
		syslog(LOG_ERR, "Failed to connect to ubus");
		return -1;
	}

	ubus_add_uloop(ctx);

	// Call function to update the global structures.
	if (read_config()) {
		syslog(LOG_ERR, "failed to read config");
		goto exit;
	}

	// Create a list of dns server IP's.
	if (init_dns_server_list() != 0) {
		syslog(LOG_ERR, "error in initialising dns server ip");
		goto exit;
	}

	// Create a thread which would send out dns query
	// for all the blacklisted urls to create a dns cache.
	pthread_t ptid;

	if (pthread_create(&ptid, NULL, &send_dns_query, NULL) != 0) {
		syslog(LOG_ERR, "pthread create failed");
		goto exit;
	}

	uloop_fd.fd = open_netlink_connection();
	uloop_fd.cb = url_filter_recv;
	if (uloop_fd_add(&uloop_fd, ULOOP_READ)) {
		syslog(LOG_ERR, "url fitler: uloop_fd_add(%d) failed", uloop_fd.fd);
		goto destroy;
	}

	// Main loop of urlfilter
	uloop_run();

destroy:
	// Destroying the queue handle and closing nfq handle.
	nfq_destroy_queue(queue_h);
	nfq_close(nfq_h);

exit:
	ubus_free(ctx);
	uloop_done();

	return 0;
}

unsigned int ipv6_header_get_L4_header_offset(struct ip6_hdr *ip6h_p, uint8_t *nextHdr_p)
{
	unsigned int ext_head_count = 8;
	const struct ip6_ext *ip_ext_p;
	unsigned int payload_offset = 0;

	*nextHdr_p = ip6h_p->ip6_nxt;
	ip_ext_p = (const struct ip6_ext *)(ip6h_p + 1);
	payload_offset = sizeof(struct ip6_hdr);

	do {
		if ((*nextHdr_p == IPPROTO_TCP) || (*nextHdr_p == IPPROTO_UDP))
			break;

		payload_offset += (ip_ext_p->ip6e_len + 1) << 3;
		*nextHdr_p = ip_ext_p->ip6e_nxt;
		ip_ext_p = (struct ip6_ext *)((uint8_t *)ip6h_p + payload_offset);
		ext_head_count--; /* at most 8 extension headers */
	} while (ext_head_count);

	return payload_offset;

}
